#lang racket
(require (for-syntax racket/struct-info))

(provide define-aug-struct)

;; stolen from define-struct
(define-for-syntax (build-name id . parts)
  (datum->syntax
   id
   (string->symbol
    (apply string-append
           (map (lambda (p)
                  (cond
                    [(syntax? p) (symbol->string (syntax-e p))]
                    [(symbol? p) (symbol->string p)]
                    [(string? p) p]))
                parts)))
   id))

;; this unifies the super's info with the sub's info, allowing
;; you to override defaults and contract info
(define-for-syntax (unify super-info sub-info)
  (append (map (λ (sup)
                 (let ([sub (assoc (car sup) sub-info)])
                   (or sub sup)))
               super-info)
          (filter (λ (sub) (not (assoc (car sub) super-info))) sub-info)))
               

;; the general idea is that, assuming sup is either another aug-struct
;; or #f,
;;
;; (define-aug-struct (sub sup)
;;   ([field1 contract1]
;;    [field2 contract2 default2]
;;    ...))
;;
;; should be expanded to
;;
;; (define-syntax sub-info
;;   '([field1 contract1] [field2 contract2 default2] ...))
;;   so I can grab this info later to unify it with any sub-structs
;; (define-struct (sub sup)
;;   (field2))        assumes field1 was defined in sup
;; (define (create-sub #:field1 field1
;;                     #:field2 [field2 default2])
;;    (make-sub field1 field2))
;; (provide/contract 
;;     [create-sub (-> #:field1 contract1
;;                     #:field2 contract2
;;                     sub?)])
(define-syntax (define-aug-struct stx)
  (syntax-case stx ()
    [(_ (id super-id) field-info-syntax)
     ; TODO: error-checking
     (let*
         ([id+maybe-super (if (syntax->datum #'super-id)
                              #'(id super-id)
                              #'id)]
          [super-info (if (syntax->datum #'super-id)
                          (syntax-local-value (build-name #'super-id #'super-id "-info"))
                          '())]
          [id-name (syntax->datum #'id)]
          [field-info (syntax->datum #'field-info-syntax)]
          [unified-info (unify super-info field-info)]
          [subfield-names (map car (filter (λ (info) 
                                             (not (assoc (car info) super-info)))
                                           field-info))]
          [all-field-names (map car unified-info)]
          [kw-names (map (λ (fld)
                           (string->keyword (symbol->string fld)))
                         all-field-names)]
          [create-id (build-name #'id "create-" id-name)]
          [no-default (gensym)]
          [defaults (map (λ (info)
                           (if (= 2 (length info))
                               no-default
                               (caddr info)))
                         unified-info)]
          [create-formals (cons create-id 
                                (foldr append '() 
                                       (map (λ (kw fld def)
                                              (if (equal? no-default def)
                                                  (list kw fld)
                                                  (list kw (list fld def))))
                                            kw-names all-field-names defaults)))]
          [contracts (map cadr unified-info)]
          [create-contract (cons '-> 
                                 (foldr append `(,(build-name #'id id-name "?"))
                                        (map (λ (kw contr)
                                               (list kw contr))
                                             kw-names contracts)))]
          [make-id (build-name #'id "make-" id-name)]
          [id-info (build-name #'id id-name "-info")])
       (with-syntax ([unified-info-syntax unified-info])
         #`(begin
             (define-syntax #,id-info (syntax->datum #'unified-info-syntax))
             (define-struct #,id+maybe-super 
               #,subfield-names
               #:transparent)
             (define #,create-formals
               (#,make-id #,@all-field-names))
             (provide/contract [#,create-id #,create-contract]))))]
    [(_ id field-info)
     #'(define-aug-struct (id #f) field-info)]))

